export var exportedVar = "exported var";
export function exportedFunction() {
  return "exported function";
}
export let exportedLet = "exported let";
export const exportedConst = "exported const";

var notExportedVar = "not exported var";
function notExportedFunction() {
  return "not exported function";
}
let notExportedLet = "not exported let";
const notExportedConst = "not exported const";
